package morrowind;
/*
 * Dialogue.java
 *
 * File created on 19 August 2007
 * Last modified 6 April 2008
 * By Andy Joel
 *
 * Permission is granted to use and modify this as you want except that:
 * (1) You must credit me as appropriate
 * (2) You may not make money from this work or anything derived from it
 * (3) Whenever or however you release this work or anything derived from it the same conditions (at a minimum)
 *       must apply.
 *
 * If you do modify the software, I would be grateful if let me know. Send a PM to "The Pixie" at the
 * Elder Scrolls forum http://www.bethsoft.com/bgsforums/
 */

/*
 * XmlHandler is a utility class I created some years ago for handling XML, when Java support for XML was pretty
 * poor. It may not be elegant, but it does work, and has been unit tested and used in some personal applications.
 * The other the XML classes are trivial.
 * Utils contains some utility methods; again, most have been unit tested and used quite a bit by me.
 * This class has NOT been extensively tested.
  */

/*
 * The dialogue export/import format is a text file. Each record is on a new line. Each field is separated by tabs,
 * and enclosed by double quotes. The fields are:
 * Pointer to previous field ID: Blank for top of list (eg topof topic or first entry for a quest)
 * Field ID (18 to 19 digit number)
 * Pointer to next field ID: Morrowind CS throws a warning if this is wrong, but actually handles it fine anyway 
 * Type: Journal, Greeting, Topic
 * Name: The name of the quest, the greeting level (eg "Greeting 1") or the topic name
 * Text: Limited to 500? characters?
 *
 *  0  Result: Script, each line separated by a literal "[CRLF]"
 *  1  Disp/Index: The index for a journal, the minimum disposition otherwise
 *  2  NPC Faction rank, as a number from 1 to 9, or 1 if not set
 *  3  NPC Sex, as Male or Female
 *  4  Player Faction rank, as a number from 1 to 9, or 1 in not set
 *  5  NPC: The ID of the NPC (or blank)
 *  6  Race (eg Dark Elf)
 *  7  Profession/Class
 *  8  Player faction
 *  9  Location
 * 10  NPC faction
 * 11  Voice filename (with path)
 * 12  Function/Variable: Sets of four fields, as many as required
 *      Type: Journal, Global, Function, Item, etc.
 *      Name: Of the journal, global, function (i.e., Choice, etc.), item
 *      Comparator: =, >=, !=, etc.
 *      Value
 * 16, 20, 24, 28, 32
 * 36  Quest flags, Q100 for title, Q010 for finished, Q001 for restart, Q000 otherwise
 *
 * The first record in a journal is the title, \t\t\t-1\t\t-1\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tQ100\t
 * The second \t\t10\t-1\t\t-1\t\t-1\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tQ000\t
 * The rule is that the title has no index number, the second has the nunber 10, and so on.
 * There are flags: Q100 indicates the title, Q010 indicates the finish (I guess at the end as they were added in later version)
 *
 * Use "" to escape "
 */



public class Dialogue {
    private static boolean DEBUG = true;
    
    private static final int RACE = 6, CLASS = 7, LOCATION = 9, FACTION = 10;
    private static final String RESULT_RETURN = "[CRLF]";
     
    private static int SIZE = 38;
    
    private static String[] typeNames = {"default", "first", "done", "finish", "finished", "sorry", };
    private static int DEFAULT= 0, FIRST = 1, DONE = 2, FINISH = 3, FINISHED = 4, SORRY = 5;
            
    private static String startIndex;
    private static Dialogue[] dialogues;
    
    private String npcId, depends;
    private Journal journal;
    private Greeting greeting;
    private Topic topic;
    private NumberHandler numberHandler;
    private boolean greetingsStarted, following;

    private java.util.Random random = new java.util.Random();
    
    /** Creates a new instance of Dialogue */
    public Dialogue(String xmlString) throws XmlHandlerException {
        npcId = XmlHandler.getValue(xmlString, "npc");
        depends = XmlHandler.getValue(xmlString, "depends", "");
        String follows = XmlHandler.getValue(xmlString, "follows", "");
        if (!follows.equals("")) {
            depends = follows;
            following = true;
        }
        journal = new Journal(XmlHandler.getElementWithTags(xmlString, "journal"));
        greeting = new Greeting(XmlHandler.getElementWithTags(xmlString, "greeting"));
        topic = new Topic(XmlHandler.getElementWithTags(xmlString, "topic"), this);
    }
 
   
    private class Entry {
        String text;
        int from, to;
        
        Entry(String xmlString) throws XmlHandlerException {
            text = XmlHandler.getElement(xmlString, "entry");
            //item = XmlHandler.getValue(xmlString, "item", "");
            from = XmlHandler.getValueAsInt(xmlString, "from", 0);
            to = XmlHandler.getValueAsInt(xmlString, "to", 100);
        }
    }
    
    private class Known {
        String text, faction, className, location, race;
        
        Known(String xmlString) throws XmlHandlerException {
            text = XmlHandler.getElement(xmlString, "known");
            faction = XmlHandler.getValue(xmlString, "faction", "");
            className = XmlHandler.getValue(xmlString, "class", "");
            location = XmlHandler.getValue(xmlString, "location", "");
            race = XmlHandler.getValue(xmlString, "race", "");
        }
    }
    
    private class Journal {
        String name, title;
        Entry[] entries;
        
        Journal(String xmlString) throws XmlHandlerException {
            name = XmlHandler.getValue(xmlString, "name");
            title = XmlHandler.getElement(xmlString, "title");
            String[] ss = XmlHandler.getElementsWithTags(xmlString, "entry");
            if (ss.length > 10) {
                throw new XmlHandlerException("Limit exceeded in Journal; max 10 entries");
            }
            entries = new Entry[ss.length];
            for (int i = 0; i < ss.length; i++) {
                entries[i] = new Entry(ss[i]);
            }
        }
        
        public String toString() {
            numberHandler = new NumberHandler("");
            JournalRecord[] records = new JournalRecord[entries.length + 1];
            records[0] = new JournalRecord(name, title);
            records[0].set(SIZE - 2, "Q100");
 
            for (int i = 0; i < entries.length; i++) {
                records[i + 1] = new JournalRecord(name, entries[i].text);
                records[i + 1].set(1, "" + (i * 10 + 10));
                records[i + 1].set(SIZE - 2, "Q000");
            }
            records[entries.length].set(1, "100");
            records[entries.length].set(SIZE - 2, "Q010");


            // Journal entries get reversed !
            StringBuffer sb = new StringBuffer();
            for (int i = entries.length; i > 0; i--) {
                sb.append(records[i].toString());
            }
            numberHandler.setLast(true);
            sb.append(records[0].toString());
            return sb.toString();
         }
        
        String getLastButOneIndex() {
            return "" + (entries.length * 10 - 10);
        }
    }
    
    // Must keep all greetings together if doing multiple quests!!!
    private class Greeting {
        Entry[] entries;
        String brushoff, done, first;
        int disp;
        
        Greeting(String xmlString) throws XmlHandlerException {
            first = XmlHandler.getElement(xmlString, "first");
            done = XmlHandler.getElement(xmlString, "done");
            brushoff = XmlHandler.getElement(xmlString, "brushoff");
            disp = XmlHandler.getValueAsInt(xmlString, "disp");
            String[] ss = XmlHandler.getElementsWithTags(xmlString, "entry");
            entries = new Entry[ss.length];
            for (int i = 0; i < ss.length; i++) {
                entries[i] = new Entry(ss[i]);
            }
        }

        public String toString() {
            if (!greetingsStarted) {
                numberHandler = new NumberHandler("" + startIndex);
                greetingsStarted = true;
            }
            StringBuffer sb = new StringBuffer();
            GreetingRecord record;
            if (done.length() > 0) {
                record = new GreetingRecord(done);
                record.set(16, "Journal", journal.name, "=", "100");
                sb.append(record.toString());
            }
            
            for (int i = 0; i < entries.length; i++) {
                record = new GreetingRecord(entries[i].text);
                record.set(12, "Journal", journal.name, ">", "" + (entries[i].from == -1 ? 0 : entries[i].from));
                record.set(16, "Journal", journal.name, "<", "" + (entries[i].to == -1 ? 100 : entries[i].to));
                sb.append(record.toString());
            }
            
            record = new GreetingRecord(first);
            if (depends.length() == 0) {
                record.set(1, "" + disp);
            }
            record.set(12, "Journal", journal.name, "=", "0");
            if (depends.length() > 0) record.set(20, "Journal", depends, "=", "100");
            sb.append(record.toString());
            
            if (depends.length() == 0) {
                record = new GreetingRecord(brushoff);
                record.set(12, "Journal", journal.name, "=", "0");
                sb.append(record.toString());
            }
            
            return sb.toString();
         }
    }
    
    private class Topic {
        String title, finish, outstanding, done, first, thanks, accept, reject;
        String requiredItem, rewardItem, dead, startId, startWith;
        String obj, objGreeting, objThanks, objRefused, objMissing, objDone;
        boolean removeRequiredItem, leaveTopicOpen;
        private boolean following;
        int requiredItemCount, rewardItemCount, disp, gold, deadCount, startWithCount;
        Entry[] entries;
        Known[] knowns;
        Dialogue dialogue;
        
        Topic(String xmlString, Dialogue dialogue) throws XmlHandlerException {
            this.dialogue =  dialogue;
            this.following = following;
            title = XmlHandler.getElement(xmlString, "title");

            startId = XmlHandler.getValue(xmlString, "startref", "");
            // If leaveTopicOpen is true, the next topic will get appended to this one (so MUST be the same name)
            leaveTopicOpen = XmlHandler.getValueAsBoolean(xmlString, "leave-open", true);
            requiredItem = XmlHandler.getValue(xmlString, "item", "");
            requiredItemCount = XmlHandler.getValueAsInt(xmlString, "item-count", requiredItem.length() == 0 ? 0 : 1);
            removeRequiredItem = XmlHandler.getValueAsBoolean(xmlString, "item-remove", true);
            rewardItem = XmlHandler.getValue(xmlString, "reward", "");
            rewardItemCount = XmlHandler.getValueAsInt(xmlString, "reward-count", rewardItem.length() == 0 ? 0 : 1);
            dead = XmlHandler.getValue(xmlString, "dead", "");
            deadCount = XmlHandler.getValueAsInt(xmlString, "dead-count", dead.length() == 0 ? 0 : 1);
            disp = XmlHandler.getValueAsInt(xmlString, "disp", 20);
            gold = XmlHandler.getValueAsInt(xmlString, "gold", 0);
            startWith = XmlHandler.getValue(xmlString, "start-with", "");
            startWithCount = XmlHandler.getValueAsInt(xmlString, "start-with-count", startWith.length() == 0 ? 0 : 1);
            obj = XmlHandler.getValue(xmlString, "obj", "");
            if (!obj.equals("") && startWith.equals("")) {
                throw new XmlHandlerException("Must have a start-with value if there is an obj value in " + title);
            }
            objGreeting = XmlHandler.getElement(xmlString, "obj-greeting",
                                     "Ah, hello %PCName. I believe you have something for me.");
            objThanks = XmlHandler.getElement(xmlString, "obj-thanks",
                                     "Thank you.");
            objRefused = XmlHandler.getElement(xmlString, "obj-refused",
                                     "Then where is it exactly?");
            objMissing = XmlHandler.getElement(xmlString, "obj-missing",
                                     "Ah, hello %PCName. I thought you were supposed to be giving me something.");
            objDone = XmlHandler.getElement(xmlString, "obj-done",
                                     "Thank you for that.");
            travelTogetherInit(XmlHandler.getElementWithTags(xmlString, "travel-together", ""));
            

            for (int i = 0; i < commonTopics.length; i++) {
                if (commonTopics[i].name.equalsIgnoreCase(title)) {
                    startId = commonTopics[i].value;
                    title = commonTopics[i].name;  // ensure same case
                }
            }
            finish = XmlHandler.getElement(xmlString, "finish", "");
            if ((requiredItemCount > 0) && (finish.length() == 0)) {
                throw new XmlHandlerException("A finish required if requiredItemCount > 0 in " + title);
            }
            outstanding = XmlHandler.getElement(xmlString, "outstanding", "");
            if ((requiredItemCount > 0) && (outstanding.length() == 0)) {
                throw new XmlHandlerException("An outstanding required if requiredItemCount > 0 in " + title);
            }
            done = XmlHandler.getElement(xmlString, "done");
            accept = XmlHandler.getElement(xmlString, "accept");
            reject = XmlHandler.getElement(xmlString, "reject");
            first = XmlHandler.getElement(xmlString, "first");
            thanks = XmlHandler.getElement(xmlString, "thanks");
            String[] ss = XmlHandler.getElementsWithTags(xmlString, "entry");
            if (ss.length == 0) {
                if (finish.length() != 0) {
                    entries = new Entry[1];
                    entries[0] = new Entry("<entry>" + finish + "</entry>");
                } else {
                    throw new XmlHandlerException("Must be at least one entry in <topic> or a finish in " + title);
                }
            } else {
                entries = new Entry[ss.length];
                for (int i = 0; i < ss.length; i++) {
                    entries[i] = new Entry(ss[i]);
                }
            }
            ss = XmlHandler.getElementsWithTags(xmlString, "known");
            knowns = new Known[ss.length];
            for (int i = 0; i < ss.length; i++) {
                knowns[i] = new Known(ss[i]);
            }
        }

        public String toString() {
            TopicRecord[] records;
            int pointer = 0;
            int total = 5 + entries.length + knowns.length;
            if (requiredItemCount > 0) total += 2;
            if (!obj.equals("")) total += 5;
            if (!escortId.equals("")) total++;
            if (dialogue.following) total++;
            records = new TopicRecord[total];

            numberHandler = new NumberHandler(startId);
            
            records[0] = new TopicDoneRecord(title, done);
            records[1] = new TopicThanksRecord(this, requiredItemCount == 0, dialogue.journal.entries.length * 10 - 10);
            if (requiredItemCount > 0) {
                records[2] = new TopicOutstandingRecord(title, outstanding);
                records[3] = new TopicFinishRecord(this);
                pointer = 4;
            } else {
                pointer = 2;
            }
            for (int i = 0; i < entries.length; i++) {
                records[pointer++] = new TopicEntryRecord(title, entries[entries.length - 1 -i]);
            }            
            records[pointer++] = new TopicAcceptRecord(title, accept);
            records[pointer++] = new TopicRejectRecord(title, reject);
            records[pointer++] = new TopicFirstRecord(title, first);
            if (dialogue.following) records[pointer++] = new TopicFollowRecord(title);

            if (!escortId.equals("")) {
                records[pointer++] = new TopicTravelTogetherRecord(title, escortId);
            }
            
            if (!obj.equals("")) {
                records[pointer++] = new TopicObjThanksRecord(title, objThanks);
                records[pointer++] = new TopicObjRefusedRecord(title, objRefused);
                records[pointer++] = new TopicObjGreetingRecord(title, objGreeting);
                records[pointer++] = new TopicObjMissingRecord(title, objMissing);
                records[pointer++] = new TopicObjDoneRecord(title, objDone);
            }
            
            for (int i = 0; i < knowns.length; i++) {
                records[pointer++] = new TopicKnownRecord(title, knowns[i]);
            }            
            
            // new topic entries get reversed !
            StringBuffer sb = new StringBuffer();
            if (startId.equals("")) {
                for (int i = records.length - 1; i > 0; i--) {
                    sb.append(records[i].toString());
                }
                if (!leaveTopicOpen) numberHandler.setLast(true);  // RefID of next gets set to zero
                sb.append(records[0].toString());
            } else {
                for (int i = 0; i < records.length-1; i++) {
                    sb.append(records[i].toString());
                }
                if (!leaveTopicOpen) numberHandler.setLast(true);  // RefID of next gets set to zero
                sb.append(records[records.length-1].toString());
            }
            if (!escortId.equals("")) {
                sb.append(travelTogether());
            }
            return sb.toString();

//            for (int i = 0; i < records.length - 1; i++) {
//                sb.append(records[i].toString());
//            }
//            numberHandler.setLast(true);
//            sb.append(records[records.length - 1].toString());
//            return sb.toString();
         }
    }
    
    String escortId, destinationId, destinationName, scriptName, raceId, genderId;
    public void travelTogetherInit(String xmlString) throws XmlHandlerException {
        escortId = XmlHandler.getValue(xmlString, "escort-id", "");
        if (escortId.equals("")) return;
        destinationName = XmlHandler.getElement(xmlString, "travel-together");
        raceId = XmlHandler.getValue(xmlString, "race-id");
        genderId = XmlHandler.getValue(xmlString, "gender-id");
        destinationId = XmlHandler.getValue(xmlString, "destination-id", destinationName);
        scriptName = XmlHandler.getValue(xmlString, "script-name", journal.name.substring(0, 2) + "_" + escortId);
    }

    public String travelTogether() {
        String s = Utils.readFileWithReturns(
                   "C:\\Program Files\\Bethesda Softworks\\Morrowind\\escort.txt");
        s = s.replaceAll("escortid", escortId);
        s = s.replaceAll("journalid", journal.name);
        s = s.replaceAll("destinationid", destinationId);
        s = s.replaceAll("destinationname", destinationName);
        s = s.replaceAll("refno1", "" + getRand());
        s = s.replaceAll("refno2", "" + getRand());
        s = s.replaceAll("refno3", "" + getRand());
        s = s.replaceAll("refno4", "" + getRand());
        s = s.replaceAll("refno5", "" + getRand());
        s = s.replaceAll("refno6", "" + getRand());
        return s;
    }
    
    public String travelTogetherScript() {
        if (escortId.equals("")) return "";
        String s1 = Utils.readFileWithReturns(
                   "C:\\Program Files\\Bethesda Softworks\\Morrowind\\escort-script.txt");
        String s2 = s1.replaceAll("escortid", escortId);
        s1 = s2.replaceAll("journalid", journal.name);
        s2 = s1.replaceAll("destinationid", destinationId);
        s1 = s2.replaceAll("destinationname", destinationName);
        s2 = s1.replaceAll("raceid", raceId);
        s1 = s2.replaceAll("genderid", genderId);
        s2 = s1.replaceAll("scriptname", scriptName);
        return s2;
    }
    
    public static void initialise() {
        try {
            String xmlString = Utils.readFile("C:\\Program Files\\Bethesda Softworks\\Morrowind\\dialogue.xml");
            if (xmlString.length() == 0) {
                System.err.println("Zero length XML string");
                System.exit(1);
            }
            startIndex = XmlHandler.getValue(xmlString, "startref");
            String[] ss = XmlHandler.getElementsWithTags(xmlString, "dialogue");
            dialogues = new Dialogue[ss.length];
            for (int i = 0; i < ss.length; i++) {
                dialogues[i] = new Dialogue(ss[i]);
            }
        } catch (XmlHandlerException ex) {
            System.out.println(ex.getMessage());
            ex.printStackTrace();
        }
    }

    public static String getJournalEntries() {
        StringBuffer output = new StringBuffer();
        for (int i = 0; i < dialogues.length; i++) {
            output.append(dialogues[i].journal.toString());
        }
        return output.toString();
    }
    
    public static String getGreetingEntries() {
        StringBuffer output = new StringBuffer();
        for (int i = 0; i < dialogues.length; i++) {
            output.append(dialogues[i].greeting.toString());
        }
        return output.toString();
    }
    
    public static String getTopicEntries() {
        StringBuffer output = new StringBuffer();
        for (int i = 0; i < dialogues.length; i++) {
            output.append(dialogues[i].topic.toString());
        }
        return output.toString();
    }
    
    public static String getScripts() {
        StringBuffer output = new StringBuffer();
        for (int i = 0; i < dialogues.length; i++) {
            output.append(dialogues[i].travelTogetherScript());
        }
        return output.toString();
    }
    
    

    private class Record {
        String name, text;
        String[] fields;

        Record(String name, String text) {
            this.name = name;
            this.text = text;
            fields = new String[SIZE];
            fields[2] = "-1";
            fields[4] = "-1";
        }

        void set(int index, String value) {
            fields[index] = value;
        }

        void set(int index, String value1, String value2, String value3, String value4) {
            fields[index] = value1;
            fields[index+1] = value2;
            fields[index+2] = value3;
            fields[index+3] = value4;
        }

        String getType() {
            return "";
        }


        public String toString() {
            StringBuffer sb = new StringBuffer(numberHandler.getIds() + "\t");
            sb.append("\"" + getType() + "\"\t\"" + name + "\"\t\"" + text + "\"");
            for (int i = 0; i < SIZE; i ++) {
                sb.append("\t" + (fields[i] == null ? "" : fields[i]));
            }
            return sb.toString() + "\n";
        }
    }

    private class JournalRecord extends Record {
        JournalRecord(String name, String text) {
            super(name, text);
        }

            
        String getType() {
            return "Journal";
        }
    }

    private class TopicRecord extends Record {
        TopicRecord(String name, String text) {
            super(name, text);
            fields[5] = npcId;
        }

        TopicRecord(Record record, String text) {
            super(record.name, text);
            fields = new String[SIZE];
            for (int i = 0; i < fields.length; i++) {
                fields[i] = record.fields[i];
            }
        }

        String getType() {
            return "Topic";
        }
        
        void unsetNpc() {
            fields[5] = "";
        }
        
        void setNpc(String npcId) {
            fields[5] = npcId;
        }
        
        /*
         * The idea is this is recursive; if the 
         */
        public String toString() {
            if (text.indexOf("~") == -1) {
                return super.toString();
            }
            if (fields[20] != null) {
                System.err.println("Cannot have info dump here");
                System.err.println(text);
                System.exit(1);
            }
            String[] ss = Utils.assembleList(text, "~");
            StringBuffer sb = new StringBuffer();
            Record record = new TopicRecord(this, ss[0]);
            record.set(0, "\"Choice \"\"Continue\"\" 1\"");
            sb.append(record.toString());
            for (int i = 1; i < ss.length - 1; i++) {
                record = new TopicRecord(this, ss[i]);
                record.set(20, "Function", "Choice", "=", "" + i);
                record.set(0, "\"Choice \"\"Continue\"\" " + (i + 1) + "\"");
                sb.append(record.toString());
            }
            record = new TopicRecord(this, ss[ss.length - 1]);
            record.set(20, "Function", "Choice", "=", "" + (ss.length - 1));
            sb.append(record.toString());
            return sb.toString();
        }
    }

    private class TopicFollowRecord extends TopicRecord {
        TopicFollowRecord(String name) {
            super(name, "I might have another job for you later.");
            set(12, "Journal", journal.name, "=", "0");
            set(16, "Journal", depends, "<", "100");
            set(20, "Journal", depends, ">", "0");
        }
    }
    
    private class TopicFirstRecord extends TopicRecord {
        TopicFirstRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "0");
            if (depends.length() > 0) set(20, "Journal", depends, "=", "100");
            set(0, "\"Choice \"\"Sure, I will do my best\"\" 101 \"\"Sorry, I am too busy\"\" 102\"");
        }
    }
    
    private class TopicRejectRecord extends TopicRecord {
        TopicRejectRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "0");
            set(20, "Function", "Choice", "=", "102");
        }
    }
    
    private class TopicAcceptRecord extends TopicRecord {
        TopicAcceptRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "0");
            set(20, "Function", "Choice", "=", "101");
            set(0, "\"Journal " + journal.name + " 10\"");
            if (topic.startWith.equals("")) {
                set(0, "\"Journal " + journal.name + " 10\"");
            } else {
                set(0, "\"Journal " + journal.name + " 10[CRLF]Player->AddItem " + topic.startWith + " " + topic.startWithCount + "\"");
            }
        }
    }
    
    private class TopicEntryRecord extends TopicRecord {
        TopicEntryRecord(String name, Entry entry) {
            super(name, entry.text);
            if (entry.to == entry.from) {
                set(12, "Journal", journal.name, "=", "" + entry.to);
            } else {
                if (entry.to == 100) {
                    set(12, "Journal", journal.name, "<", "100");
                } else {
                    set(12, "Journal", journal.name, "<=", "" + entry.to);
                }
                if (entry.from == 0) {
                    set(16, "Journal", journal.name, ">", "0");
                } else {
                    set(16, "Journal", journal.name, ">=", "" + entry.from);
                }
            }
        }
    }
    
    private class TopicFinishRecord extends TopicRecord {
        TopicFinishRecord(Topic topic) {
            super(topic.title, topic.finish);
            set(12, "Journal", journal.name, "<", "100");
            set(16, "Journal", journal.name, ">", "0");
            set(0, "\"Choice \"\"Sure, here you go\"\" 101 \"\"Sorry, not yet\"\" 102\"");
            if (!topic.requiredItem.equals("")) {
                set(20, "Item", topic.requiredItem, ">=", "" + topic.requiredItemCount);
            }
            if (!topic.dead.equals("")) {
                set(24, "Dead", topic.dead, ">=", "" + topic.deadCount);
            }
        }
    }
    
    private class TopicOutstandingRecord extends TopicRecord {
        TopicOutstandingRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "<", "100");
            set(16, "Journal", journal.name, ">", "0");
            set(20, "Function", "Choice", "=", "102");
        }
    }

    private class TopicKnownRecord extends TopicRecord {
        TopicKnownRecord(String name, Known known) {
            super(name, known.text);
            set(12, "Journal", journal.name, "<", "100");
            set(16, "Journal", journal.name, ">", "0");
            if (!known.race.equals("")) set(RACE, known.race);
            if (!known.className.equals("")) set(CLASS, known.className);
            if (!known.location.equals("")) set(LOCATION, known.location);
            if (!known.faction.equals("")) set(FACTION, known.faction);
            unsetNpc();
        }
    }

    
    
    
    private class TopicThanksRecord extends TopicRecord {
        TopicThanksRecord(Topic topic, boolean noRequiredObject, int targetIndex) {
            super(topic.title, topic.thanks);
            StringBuffer sb = new StringBuffer("Journal " + journal.name + " 100");
            if (topic.disp > 0) sb.append("[CRLF]ModDisposition " + topic.disp);
            if (topic.gold > 0) sb.append("[CRLF]Player->AddItem gold_001 " + topic.gold);
            if (topic.rewardItemCount > 0) sb.append("[CRLF]Player->AddItem " + topic.rewardItem + " " + topic.rewardItemCount);
            if (topic.removeRequiredItem && !topic.requiredItem.equals("")) {
                sb.append("[CRLF]Player->RemoveItem " + topic.requiredItem + " " + topic.requiredItemCount);
            }
            set(12, "Journal", journal.name, "=", "" + targetIndex);
            if (!noRequiredObject) {
                set(16, "Function", "Choice", "=", "101");
            }
            set(0, "\"" + sb.toString() + "\"");
        }
    }
    
    private class TopicDoneRecord extends TopicRecord {
        TopicDoneRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "100");
        }
    }
    
    private class TopicObjRecord extends TopicRecord {
        TopicObjRecord(String name, String text) {
            super(name, text);
            setNpc(topic.obj);
        }
    }
    
    private class TopicTravelTogetherRecord extends TopicRecord {
        TopicTravelTogetherRecord(String name, String escortid) {
            super(name, "Shall we travel together?");
            setNpc(escortid);
            set(12, "Journal", journal.name, "<", "30");
        }
    }
    
    /*
     * journal =
     * 10 quest given
     * 20 
     *
     */
    private class TopicObjGreetingRecord extends TopicObjRecord {
        TopicObjGreetingRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "10");
            set(0, "\"Choice \"\"Sure, here you go\"\" 101 \"\"Sorry, I haven't\"\" 102\"");
            set(16, "Item", topic.startWith, ">=", "" + topic.startWithCount);
        }
    }
    
    private class TopicObjMissingRecord extends TopicObjRecord {
        TopicObjMissingRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "10");
            set(16, "Item", topic.startWith, "<", "" + topic.startWithCount);
        }
    }
    
    private class TopicObjRefusedRecord extends TopicObjRecord {
        TopicObjRefusedRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, "=", "10");
            set(16, "Function", "Choice", "=", "102");
        }
    }

    private class TopicObjThanksRecord extends TopicObjRecord {
        TopicObjThanksRecord(String name, String text) {
            super(name, text);
            StringBuffer sb = new StringBuffer("Journal " + journal.name + " 20");
            sb.append("[CRLF]ModDisposition 20");
            sb.append("[CRLF]Player->RemoveItem " + topic.startWith + " " + topic.startWithCount);
            set(0, "\"" + sb.toString() + "\"");
            set(12, "Journal", journal.name, "=", "10");
            set(16, "Function", "Choice", "=", "101");
        }
    }
    
    private class TopicObjDoneRecord extends TopicObjRecord {
        TopicObjDoneRecord(String name, String text) {
            super(name, text);
            set(12, "Journal", journal.name, ">", "10");
            set(16, "Journal", journal.name, "<", "100");
        }
    }
    
    
    
    
    
    private class GreetingRecord extends Record {
        GreetingRecord(String text) {
            super("Greeting 5", text);
            fields[5] = npcId;
        }

        String getType() {
            return "Greeting";
        }
    }

    
    private class NumberHandler {
        long previousId, thisId, nextId;
        boolean last;
        String previousAsString;
        
        NumberHandler(String s) {
            previousAsString = s;
            thisId = getRand();
            nextId = getRand();
        }
        
        void setLast(boolean last) {
            this.last = last;
        }
        
        String getIds() {
            String s;
            if (last) {       
                s = "\"" + (previousAsString == null ? "" + previousId : previousAsString)
                     + "\"\t\"" + thisId + "\"\t\"\"";
            } else {
                s = "\"" + (previousAsString == null ? "" + previousId : previousAsString)
                     + "\"\t\"" + thisId + "\"\t\"" + nextId + "\"";
            }
            previousId = thisId;
            thisId = nextId;
            nextId = getRand();
            previousAsString = null;
            return s;
        }

    }
    
    private long getRand() {
        long ran = random.nextLong();
        return ran > 0 ? ran : -ran;
    }
    
    
    private void checkLength(String s) {
        if (s.length() > 500) {
            System.err.println("========  TEXT TOO LONG  =======\n" + s);
            System.exit(1);
        }
    }
    
    private nameValue[] commonTopics = {
        new nameValue("824626281811019302", "Requirements"),
        new nameValue("26003290602717310479", "Advancement"),
        new nameValue("9654141542779126358", "business"),
        new nameValue("1506122032762213967", "Orders"),
        new nameValue("1729913769476329813", "duties"),
        new nameValue("22938157822342711483", "chores"),
        new nameValue("5110146922059517766", "rules"),
        new nameValue("22761111983241720411", "jobs"),
        new nameValue("951829803251657388", "work"),
    };
    
    
    private class nameValue{
        String name;
        String value;
        nameValue(String value, String name) {
            this.value = value;
            this.name = name;
        }
    }
    
    public static void main(String[] args) {
        initialise();
        Utils.writeFile("C:\\Program Files\\Bethesda Softworks\\Morrowind\\scripts.txt", getScripts());
        StringBuffer sb = new StringBuffer();
        sb.append(getJournalEntries());
        sb.append(getGreetingEntries());
        sb.append(getTopicEntries());
        Utils.writeFile("C:\\Program Files\\Bethesda Softworks\\Morrowind\\dialogue.txt", sb.substring(0, sb.length()-1));
        System.out.println("Done.");
    }
    
}
