package morrowind;

/**
 * A set of static methods for use with String objects.
 */

public class Utils {

    private Utils() {
    }

  /**
   * Returns string in which any occurance of a second string is replaced by a third string.
   *
   * @param source  The source string
   * @param was     The substring to be replaced
   * @param now     The replacement string
   * @return        The resultant string
   * @throws NullPointerException if any parameter is null
   * @throws IllegalArgumentException if 'was' is zero length or is a substring of 'now'
   */
  public static String tokenSwap(String source, String was, String now) {
    if (was.length() == 0)
      throw new java.lang.IllegalArgumentException("Cannot have was=\"\" in pixie.util.StringMaker.tokenSwap");
    if (now.indexOf(was) != -1)
      throw new java.lang.IllegalArgumentException("Cannot have was as a substring of now in pixie.util.StringMaker.tokenSwap");
    while (source.indexOf(was) != -1)
      source = source.substring(0, source.indexOf(was))
               + now
               + source.substring(source.indexOf(was) + was.length());
    return source;
  }

  /**
   * Returns as a string the contents of the specified file.
   * Line breaks are removed.
   *
   * @param   filename      The name of the file.
   * @return                The file contents.
   */
  public static String readFile(String filename) {
      return readFile(filename, null, false);
  }


  /**
   * Returns as a string the contents of the specified file.
   * Line breaks are removed.
   * This can be called by an applet to read a file from the document base
   * using a URL connection. If applet in null the file is opened on the local
   * drive as a file stream.
   *
   * @param   filename      The name of the file.
   * @param   applet        The calling applet (or null).
   * @return                The file contents.
   */
  public static String readFile(String filename, java.applet.Applet applet) {
    return readFile(filename, applet, false);
  }


  /**
   * Returns as a string the contents of the specified file.
   * Line breaks are preserved.
   *
   * @param   filename      The name of the file.
   * @return                The file contents.
   */
  public static String readFileWithReturns(String filename) {
    return readFile(filename, null, true);
  }

  /**
   * Returns as a string the contents of the specified file.
   * Line breaks are removed if keepReturns is false.
   * This can be called by an applet to read a file from the document base
   * using a URL connection. If applet in null the file is opened on the local
   * drive as a file stream.
   *
   * @param   filename      The name of the file.
   * @param   applet        The calling applet (or null).
   * @param   keepReturns    End of line characters are converted to spaces if false.
   * @return                The file contents.
   */
  public static String readFile(String filename, java.applet.Applet applet, boolean keepReturns) {
    try {
      if (applet != null) {
        java.net.URL url = new java.net.URL(applet.getDocumentBase(), filename);
        return readFile(url.openStream(), keepReturns);
      } else {
        return readFile(new java.io.FileInputStream(filename), keepReturns);
      }
    } catch (Exception ex) { // expected: MalformedURL, IO and FileNotFound
      System.err.println("Caught Exception: " + ex);
      return "";
    }
  }






  /**
   * Returns as a string the contents of the specified <code>java.io.InputStream</code>.
   * Line breaks are removed if keepReturns is false, and the <code>java.io.InputStream</code> is closed
   * when the end of the file is reached. All other readFile methods use this.
   * If a problem is encountered a message is sent to System.err and a zero-length String is returned.
   * A good way to use this method is:
   * <pre>
   * String s = StringMaker.readFile(getClass().getResourceAsStream(filename), false);
   * </pre>
   *
   * @param   is             The stream to take the String from.
   * @param   keepReturns    End of line characters are converted to spaces if false.
   * @return                 The file contents.
   */
  public static String readFile(java.io.InputStream is, boolean keepReturns) {
    java.io.BufferedReader in;
    StringBuffer sb = new StringBuffer();
    char terminator = keepReturns ? '\n' : ' ';
    try {
      in = new java.io.BufferedReader(new java.io.InputStreamReader(is));
      try {
        String s = in.readLine();
        sb = new StringBuffer();
        while (s != null) {
          sb.append(s);
          sb.append(terminator);
          s = in.readLine();
        }
        in.close();
      } catch (NullPointerException e) {
        System.err.println("Caught NullPointerException: " + e.getMessage());
        in.close();
      }
    } catch (java.io.IOException e) {
      System.err.println("Caught IOException: " + e.getMessage());
    }
    return sb.toString();
  }




  /**
   * Returns a string with any section that starts and ends as specified
   * removed. Thus
   * <code>deleteSection(xmlString, "<!--", "-->")</code>
   * would remove any XML or HTML comments from a string.
   *
   * @param   s           The source string
   * @param   start       Section start delimiter
   * @param   end         Section end delimiter
   * @return              The array items listed.
   */
  public static String deleteSection(String s, String start, String end) {
    if (start.length() == 0)
      throw new java.lang.IllegalArgumentException("Cannot have start=\"\" in pixie.util.StringMaker.tokenSwap");
    if (end.length() == 0)
      throw new java.lang.IllegalArgumentException("Cannot have end=\"\" in pixie.util.StringMaker.tokenSwap");
    StringBuffer sb = new StringBuffer();
    int i = 0;
    while (s.indexOf(start, i) != -1) {
      sb.append(s.substring(i, s.indexOf(start, i)));
      i = s.indexOf(end, i) + end.length();
    }
    sb.append(s.substring(i, s.length()));
    return sb.toString();
  }

    /**
     * Converts a string to a string array. The given string is broken up at any white space.
     *
     * @param s  String to convert.
     * @return   The given string broken up into an array of strings.
     * @throws NullPointerException if the parameter is null
     */
    public static String[] assembleList(String s) {
        return assembleList(s, " \n\t");
    }

    /**
     * Converts a string to a string array.
     *
     * @param s  String to convert.
     * @param sep The string is broken at any occurance of any of these characters
     * @return   The given string broken up into an array of strings.
     * @throws NullPointerException if either parameter is null
     */
    public static String[] assembleList(String s, String sep) {
        java.util.Vector v = new java.util.Vector();
        java.util.StringTokenizer st = new java.util.StringTokenizer(s, sep);
        for (int i = 0; st.hasMoreTokens(); i++)
            v.add(st.nextToken());
        return getStringArray(v);
    }

  /**
   * Converts a vector to an array of strings.
   *
   * @param    v         The source of the strings.
   * @return             An array of strings derived from v.
   */
  public static String[] getStringArray(java.util.Vector v) {
    String ss[] = new String[v.size()];
    for (int i = 0; i < v.size(); i++)
      ss[i] = v.elementAt(i).toString();
    return ss;
  }

    /**
     * Creates a file using the given filename and text. Any '\n' characters
     * in the text are extracted, and instead the <code>java.io.BufferedWriter.newLine()</code>
     * method is called.
     *
     * @param filemame  File to be copied
     * @param from      Source directory
     * @param to        Destination directory
     * @return          An error code, 0 indicates success, 1 a failure to open the file,
     *                  2 a failure during writing, and 3 a failure during closing.
     */
    public static int writeFile(String filename, String text) {
        java.io.BufferedWriter out = null;
        try {
            out = new java.io.BufferedWriter(new java.io.FileWriter(filename));
        } catch (java.io.IOException ex) {
            return 1;
        }
        try {
            int start = 0, end = text.indexOf('\n');
            while (end != -1) {
                out.write(text.substring(start, end));
                out.newLine();
                start = end + 1;
                end = text.indexOf('\n', start);
            }
            out.write(text.substring(start));
            out.newLine();
        } catch (java.io.IOException ex) {
            return 2;
        }
        try {
            out.close();
        } catch (java.io.IOException ex) {
            return 3;
        }
        return 0;
    }

    
}