package morrowind;

/**
 * XmlHandler offers a number of static methods that extract data from XML strings
 * and readers.
 *
 * <p>
 * XmlHandler allows certain XML documents to be interpreted. It cannot readily cope with
 * mixed content elements (elements containing both text and elements). It ignores
 * the DTD or schema; the XML structure must be known to the calling program.
 * XmlHandler ignores the XML header and comments. It assumes documents are well formated;
 * check, for example by viewing in Internet Explorer. Sending null or empty strings to
 * these functions may give unpredictable results.
 *
 * <p>
 * While I have used XmlHandler extensively, it must be admitted that I may construct my XML
 * documents in a certain way; it is quite likely that I am unconsciously using
 * a restricted form of XML.
 *
 * <p>
 * The first eight methods allow reading and writing to file.
 *
 * <p>
 * Next there are a variety of getValue methods that will retrieve the value from
 * a name-value pair in the starting tag:
 * <code>&lt;tagname name="value"&gt;Text&lt;/tagname&gt;
 * The methods all come in two flavours, one throws an XmlValueNotFoundException if the value is not
 * found, the other takes a default value and will return it. A pair of methods exists for
 * each basic variable type; java.lang.String, int, long, double, boolean (accepts 'true', 'false',
 * 'yes' and 'no') and java.awt.Color. Also a pair of methods exists that will compare the value
 * against an array of Strings and returns the position in the array. All these
 * methods throw XmlHandlerException if a problem is encountered with the XML,
 * NullPointerException if the name or XML source is null or IllegalArgumentException
 * if the name has a length of zero.
 *
 * <p>
 * There are a variety of getElement methods that will retrieve elements from a String
 * or java.io.Reader.
 * The methods come in two flavours, one will return the element with the tags
 * (eg getElementWithTags) and one without (eg getElement).
 * These should be used carefully; an attempt to extract values from a tag after a call
 * to getElement rather than getElementWithTags is doomed to failure.
 * A second variation allows one to extract a single element or a set (with identical tags),
 * and if extracting a single element, a default value can be supplied which will
 * be used if the element is not found.
 * Finally the element can be extracted from a String or a java.io.Reader.
 * All these
 * methods throw XmlHandlerException if a problem is encountered with the XML,
 * NullPointerException if the tag or XML source is null or IllegalArgumentException
 * if the tag has a length of zero.
 *
 * <p>
 * This class can handle empty XML elements in the form <code>&lt;tag/&gt;</code>,
 * but will expand it into the form <code>&lt;tag&gt;&lt;/tag&gt;</code>.
 *
 * <p>
 * Revision 1.2 causes an XmlValueNotFoundException to be returned if an element of value is not
 * found. Methods with default values now only return the default value if this
 * exception is thrown, and not if the XML is invalid. This means that pretty much every
 * method now throws XmlHandlerException, as they can all encounter bad XML.
 *
 * @author Andy Joel
 * @version 1.2 16/April/04
 */



public class XmlHandler {

  private XmlHandler() {
  }

  /**
   * Returns as a <code>java.io.BufferedReader</code>, opened using the given
   * filename. If this is used from an applet, the applet can be sent as a parameter, and
   * the file will be opened using <code>java.net.URL.openStream()</code>;
   * otherwise, if applet is <code>null</code> the file is opened from the standard
   * file system.
   *
   * @param   filename      The name of the file
   * @param   applet        The calling applet or null
   * @return                File as a <code>java.io.BufferedReader</code>
   */
  public static java.io.BufferedReader openFile(String filename, java.applet.Applet applet)
    throws java.io.FileNotFoundException, java.io.IOException, java.net.MalformedURLException {
    java.io.BufferedReader in;
    if (applet != null) {
      java.net.URL url = new java.net.URL(applet.getDocumentBase(), filename);
      return new java.io.BufferedReader(new java.io.InputStreamReader(url.openStream()));
    } else {
      return new java.io.BufferedReader(new java.io.FileReader(filename));
    }
  }



//   ******************   READING FROM FILE   ***********************

  /**
   * Returns as a string the contents of the specified file with whitespace
   * collapsed and comments removed (ie with just the XML text remaining).
   *
   * @param   filename      The name of the file
   * @return                The file contents as a single line of text
   */
  public static String readFile(String filename) {
    return readFile(filename, (java.applet.Applet)null);
  }


  /**
   * Returns as a string the contents of the specified file.
   * This can be called by an applet to read a file from the document base
   * using a URL connection. If applet in null the file is opened on the local
   * drive as a file stream.
   *
   * @param   filename      The name of the file.
   * @param   applet        The calling applet (or null).
   * @return                The file contents.
   */
  public static String readFile(String filename, java.applet.Applet applet) {
    try {
      if (applet != null) {
        java.net.URL url = new java.net.URL(applet.getDocumentBase(), filename);
        return readFile(url.openStream());
      } else {
        return readFile(new java.io.FileInputStream(filename));
      }
    } catch (Exception ex) { // expected: MalformedURL, IO and FileNotFound
      System.err.println("Caught Exception: " + ex);
      return "";
    }
  }

  /**
   * Returns as a string the contents of the specified file.
   *
   * @param   file      The file.
   * @return                The file contents.
   */
  public static String readFile(java.io.File file) {
    try {
      return readFile(new java.io.FileInputStream(file));
    } catch (Exception ex) { // expected: MalformedURL, IO and FileNotFound
      System.err.println("Caught Exception: " + ex);
      return "";
    }
  }


  /**
   * Returns as a string the contents of the specified stream with whitespace
   * collapsed and comments removed.
   * All the readFile methods use this, which in tiurn uses the
   * readFile(java.io.InputStream, boolean) method in pixie.util.StringMaker.
   * A useful way to use this method is:
   * <pre>
   * String s = XmlHandler.readFile(getClass().getResourceAsStream(xmlfilename));
   * </pre>
   *
   * @param   is            The InputStream
   * @return                The file contents as a single line of text
   */
  public static String readFile(java.io.InputStream is) {
    return stringWithWhitespaceCollapsed(
        stringWithoutComments(Utils.readFile(is, true)));
  }





  /**
   * Sets up the XmlUser object with values from the specified file,
   * using the fromXml() method.
   *
   * @param   filename      The name of the file
   * @param   xmlUser       The object that will be initialised by the XML file
   */
  public static void readFile(String filename, XmlUser xmlUser) throws XmlHandlerException {
    xmlUser.fromXml(readFile(filename, (java.applet.Applet)null));
  }

  /**
   * Sets up the XmlUser object with values from the specified file,
   * using the fromXml() method.
   *
   * @param   filename      The name of the file
   * @param   xmlUser       The object that will be initialised by the XML file
   * @param   applet        The calling applet (or null)
   */
  public static void readFile(String filename, XmlUser xmlUser, java.applet.Applet applet) throws XmlHandlerException {
    xmlUser.fromXml(readFile(filename, applet));
  }



//   ******************   WRITING TO FILE   ***********************


  /**
   * Writes the XmlUser object values to the specified file,
   * using the toXml() method. It renames the old file to the <code>filename + ".bak"</code>
   * first, as a back up (this may be lost on a second unsuccessful attempt).
   *
   * @param   filename      The name of the file (without ".xml" extension)
   * @param   xmlUser       The object that will be initialised by the XML file
   */
  public static boolean writeFile(XmlUser xmlUser, String filename) {
    return writeFile(xmlUser, filename, null);
  }

  /**
   * Writes the XmlUser object values to the specified file,
   * using the toXml() method. It renames the old file to the <code>filename + ".bak"</code>
   * first, as a back up (this may be lost on a second unsuccessful attempt).
   *
   * @param   filename      The name of the file (without ".xml" extension)
   * @param   xmlUser       The object that will be initialised by the XML file
   * @param   directory     The directory of the file
   */
  public static boolean writeFile(XmlUser xmlUser, String filename, String directory) {
        String dir = (directory == null ? "" : directory + java.io.File.separator);
        java.io.File file = new java.io.File(dir + filename + ".xml");
        file.renameTo(new java.io.File(dir + filename + ".bak"));
        try {
            java.io.PrintWriter out = new java.io.PrintWriter(
                new java.io.FileOutputStream(dir + filename + ".xml"));
            out.println(xmlUser.toXml());
            out.close();
            return true;
        } catch (java.io.FileNotFoundException ex) {
            System.err.println("ERROR in putDataToFile: " + ex);
            return false;
        }
    }



//  ----------------  getValue methods  --------------




  /**
    * Allows the value to be taken from a name-value pair in the string.
    * For example, from the string
    * <code>&lt;tag myname="andy"&gt;Some text&lt;/tag&gt;</code>
    * "andy" will be returned using "myname" as the name parameter.
    * The name-value pair must appear in the first tag in the string.
    * This is the generic getValue method. It uses the class of defaultValue
    * to decide how to handle the value, so defaultValue must be present,
    * even if it is not to be used.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue The value to return if none is found. The returned object will be of this type.
    * @param   useDefault   If true, defaultValue is returned if the name-value pair
    *                       is not found, otherwise an XmlValueNotFoundException is thrown.
    * @return               The value found or defaultValue otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static Object getValue(String xmlString, String name,
                                Object defaultValue, boolean useDefault)
                                   throws XmlHandlerException {
    String s;
    try {
      s = getValue(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      if (useDefault)
        return defaultValue;
      else
        throw e;
    }
    if (defaultValue instanceof Integer)
      return new Integer(s);
    if (defaultValue instanceof Long)
      return new Long(s);
    if (defaultValue instanceof Boolean)
      return new Boolean(s);
    if (defaultValue instanceof String)
      return s;
    if (defaultValue instanceof Double)
      return new Double(s);
    throw new java.lang.IllegalArgumentException("defaultValue class " + defaultValue.getClass() + " not recognised.");
  }


  /**
    * Allows the value to be taken from a name-value pair in the string.
    * For example, from the string
    * <code>&lt;tag myname="andy"&gt;Some text&lt;/tag&gt;</code>
    * "andy" will be returned using "myname" as the name parameter.
    * The name-value pair must appear in the first tag in the string.
    * This is the basic getValue method; other provide some handling
    * of the resultant string.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue The value to return if none is found
    * @return               The value found or defaultValue otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String getValue(String xmlString, String name,
                                String defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValue(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }

  /**
    * Allows the value to be taken from a name-value pair in the string.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @return               The value found.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    * @exception  NullPointerException If the name or xmlString is null.
    * @exception  IllegalArgumentException If the name is zero length.
    */
  public static String getValue(String xmlString, String name)
                        throws XmlHandlerException {

    if (name == null)
      throw new java.lang.NullPointerException("name = null in XmlHandler");
    if (xmlString == null)
      throw new java.lang.NullPointerException("xmlString = null in XmlHandler");
    if (name.length() == 0)
      throw new java.lang.IllegalArgumentException("Zero length name found by XmlHandler");

    boolean okay = false;
    int pos = 0, pos2 = 0, pos3 = 0;
    while (!okay) {
      okay = true;
      pos = xmlString.indexOf(name, pos + 1);
      if (pos == -1)
        throw new XmlValueNotFoundException("Looking for " + name);

      int posGt = xmlString.indexOf('>');
      if (posGt == -1)
        throw new XmlHandlerException("Looking for " + name + ", failed to find \'>\'",
                          XmlHandlerException.BAD_XML);
      if (pos > posGt)
        throw new XmlValueNotFoundException("Looking for " + name + ", failed to find before \'>\'");

      if (!Character.isWhitespace(xmlString.charAt(pos - 1))) { okay = false; continue; }
      pos2 = pos + name.length();
      if (Character.isLetterOrDigit(xmlString.charAt(pos2))) { okay = false; continue; }

      while (Character.isWhitespace(xmlString.charAt(pos2))) pos2++;
      if (xmlString.charAt(pos2) != '=')
        throw new XmlHandlerException("Looking for " + name + ", no equals found after " + xmlString.substring(0, pos2),
                          XmlHandlerException.BAD_XML);
      pos2++;
      while (Character.isWhitespace(xmlString.charAt(pos2))) pos2++;
      if (xmlString.charAt(pos2) != '\"')
        throw new XmlHandlerException("Looking for " + name + ", no start quote found",
                          XmlHandlerException.BAD_XML);
      pos2++;
      pos3 = xmlString.indexOf('\"', pos2);
      if (pos3 == -1)
        throw new XmlHandlerException("Looking for " + name + ", no end quote found",
                          XmlHandlerException.BAD_XML);
    }

    return xmlString.substring(pos2, pos3);
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as a boolean.
    * If the name is not found or value cannot be converted to a boolean,
    * an exception is thrown. The only valid values are true, false,
    * yes and no (all lower case).
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @return               The value found or false otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    */
  public static boolean getValueAsBoolean(String xmlString, String name)
                    throws XmlHandlerException {
    String s = getValue(xmlString, name);
    if (s.equals("yes") || s.equals("true"))
      return true;
    else if (s.equals("no") || s.equals("false"))
      return false;
    else
      throw new XmlHandlerException("Found " + s + " for boolean for " + name,
                    XmlHandlerException.BAD_ARGUMENT);
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as a boolean.
    * If the name is not found or value cannot be converted to a boolean,
    * the default is returned. The only valid values are true, false, yes and no.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue Value to use if none is found
    * @return               The value found or def otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static boolean getValueAsBoolean(String xmlString, String name,
                                          boolean defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValueAsBoolean(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as a long.
    * If the name is not found or value cannot be converted to a number,
    * an exception is thrown.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @return               The value found or zero otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    */
  public static long getValueAsLong(String xmlString, String name)
                        throws XmlHandlerException {
    String str = getValue(xmlString, name);
    try {
      return Long.parseLong(str);
    } catch (NumberFormatException e) {
      throw new XmlHandlerException("Could not parse " + str + " as long for " + name,
                        XmlHandlerException.BAD_ARGUMENT);
    }
  }

  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as a long.
    * If the name is not found or value cannot be converted to a long,
    * the default is returned.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue Value to use if none is found
    * @return               The value found or def otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static long getValueAsLong(String xmlString, String name,
                        long defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValueAsLong(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as an int; the value should be written in binary.
    * If the name is not found or value cannot be converted to a number,
    * an exception is thrown.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @return               The value found or zero otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    */
  public static long getValueAsBinary(String xmlString, String name)
                        throws XmlHandlerException {
    String str = getValue(xmlString, name);
    try {
      return Long.parseLong(str, 2);
    } catch (NumberFormatException e) {
      throw new XmlHandlerException("Could not parse " + str + " as long for " + name,
                        XmlHandlerException.BAD_ARGUMENT);
    }
  }

  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as an int; the value should be written in binary.
    * If the name is not found,
    * the default is returned.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue Value to use if none is found
    * @return               The value found or def otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static long getValueAsBinary(String xmlString, String name,
                        long defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValueAsBinary(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as an int.
    * If the name is not found or value cannot be converted to a number,
    * an exception is thrown.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @return               The value found or zero otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    */
  public static int getValueAsInt(String xmlString, String name)
                        throws XmlHandlerException {
    String str = getValue(xmlString, name);
    try {
      return str.startsWith("+") ? Integer.parseInt(str.substring(1)) : Integer.parseInt(str);
    } catch (NumberFormatException e) {
      throw new XmlHandlerException("Could not parse " + str + " as int for " + name,
                        XmlHandlerException.BAD_ARGUMENT);
    }
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as an int.
    * If the name is not found or value cannot be converted to a number,
    * defaultValue is returned.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue The default value.
    * @return               The value found or the default otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static int getValueAsInt(String xmlString, String name,
                        int defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValueAsInt(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }





  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as an int.
    * The value is compared with the result of the toString method on each object
    * in the array list, and the index
    * If the name is not found an exception is thrown.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   list         List of acceptable string values
    * @return               The array index found or zero otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    */
  public static int getValueAsInt(String xmlString, String name,
                                  Object[] list)
                        throws XmlHandlerException {
    String s = getValue(xmlString, name);
    for (int i = 0; i < list.length; i++)
      if (s.equals(list[i].toString()))
        return i;
    throw new XmlHandlerException("Argument is " + s + " for " + name,
                        XmlHandlerException.BAD_ARGUMENT);
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as an int.
    * The value is compared with the result of the toString method on each object
    * in the array list, and the index
    * is returned.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   list         List of acceptable string values
    * @param   defaultValue The default value.
    * @return               The array index found or the default otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static int getValueAsInt(String xmlString, String name,
                                  Object[] list, int defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValueAsInt(xmlString, name, list);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as a double.
    * If the name is not found or value cannot be converted to a number,
    * an exception is thrown.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @return               The value found or zero otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the value is not found.
    */
  public static double getValueAsDouble(String xmlString, String name)
                        throws XmlHandlerException {
    String str = getValue(xmlString, name);
    try {
      return Double.valueOf(str).doubleValue();
    } catch (NumberFormatException e) {
      throw new XmlHandlerException("Could not parse " + str + " as double for " + name,
                        XmlHandlerException.BAD_ARGUMENT);
    }
  }


  /**
    * Allows the value to be taken from a name-value pair in the string,
    * returning it as a double.
    * If the name is not found or value cannot be converted to a number,
    * defaultValue is returned.
    *
    * @param   xmlString    The XML text to examine.
    * @param   name         The name of the name-value pair.
    * @param   defaultValue The default value.
    * @return               The value found or the default otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static double getValueAsDouble(String xmlString, String name,
                        double defaultValue)
                                   throws XmlHandlerException {
    try {
      return getValueAsDouble(xmlString, name);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }







//  -------------- getElement methods ---------------------



  /**
    * Allows the text to be taken from an XML element.
    * For example, from the string
    * <code>&lt;tag myname="andy"&gt;Some text&lt;/tag&gt;</code>
    * "Some text" will be returned using "tag" as the tag parameter.
    * All text between the start and end tags, including any contained
    * elements, will be returned. The tags can be anywhere in the string,
    * but the start tag must be before the end tag and both must be present.
    *
    * @param   xmlString   The XML text to examine.
    * @param   tag         The tag to look for (without &lt; and &gt;).
    * @return              The element found or the empty string ("") otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the element is not found.
    */
  public static String getElement(String xmlString, String tag)
                        throws XmlHandlerException {
    return stringWithoutTags(getElementWithTags(xmlString, tag));
  }

  /**
    * Allows the text to be taken from an XML element.
    *
    * @param   xmlString    The XML text to examine.
    * @param   tag          The tag to look for.
    * @param   defaultValue The default value.
    * @return               The element found or the default otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String getElement(String s, String tag,
                                          String defaultValue)
                        throws XmlHandlerException {
    try {
      return getElement(s, tag);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }

  /**
    * Allows the text to be taken from an XML element.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    *
    * @param   xmlString    The XML text to examine.
    * @param   tag          The tag to look for.
    * @return               The element found or the empty string ("") otherwise.
    * @exception  XmlHandlerException If the value is not found or is not valid.
    */
  public static String getElementWithTags(String xmlString, String tag)
                        throws XmlHandlerException {
    java.io.Reader in;
    try {
      in = new java.io.StringReader(xmlString);
      String s = XmlHandler.getElementWithTags(in, tag);
      in.close();
      return s;
    } catch (java.io.IOException e) {
      throw new XmlHandlerException("Problem with StringReader: " + e,
                                    XmlHandlerException.BAD_ARGUMENT);
    }
  }

  /**
    * Allows the text to be taken from an XML element.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    *
    * @param   xmlString    The XML text to examine.
    * @param   tag          The tag to look for.
    * @param   defaultValue The default value.
    * @return               The element found or the default otherwise.
    */
  public static String getElementWithTags(String s, String tag,
                                          String defaultValue)
                        throws XmlHandlerException {
    try {
      return getElementWithTags(s, tag);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }


  /**
    * Allows the text to be taken from an XML element.
    * Multiple instances of the element are collected in a string array.
    * A zero length array is returned if no elements are found.
    *
    * @param   xmlString    The XML text to examine.
    * @param   tag          The tag to look for.
    * @return               An array of the elements found.
    * @exception  XmlHandlerException If the XML is not valid.
    */
  public static String[] getElements(String xmlString, String tag)
                        throws XmlHandlerException {
    java.io.BufferedReader in;
    try {
      in = new java.io.BufferedReader(new java.io.StringReader(xmlString));
      String[] ss = XmlHandler.getElements(in, tag);
      in.close();
      return ss;
    } catch (java.io.IOException e) {
      throw new XmlHandlerException("Problem with StringReader: " + e,
                                    XmlHandlerException.BAD_ARGUMENT);
    }
  }

  /**
    * Allows the text to be taken from an XML element.
    * Multiple instances of the element are collected in a string array.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    * A zero length array is returned if no elements are found.
    *
    * @param   xmlString    The XML text to examine.
    * @param   tag          The tag to look for.
    * @return               An array of the elements found.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String[] getElementsWithTags(String xmlString, String tag)
                        throws XmlHandlerException {
    java.io.BufferedReader in;
    try {
      in = new java.io.BufferedReader(new java.io.StringReader(xmlString));
      String[] ss = XmlHandler.getElementsWithTags(in, tag);
      in.close();
      return ss;
    } catch (java.io.IOException e) {
      throw new XmlHandlerException("Problem with StringReader: " + e,
                                    XmlHandlerException.BAD_ARGUMENT);
    }
  }




  /**
    * Allows the text to be taken from a <code>java.io.Reader</code>.
    * Multiple instances of the element are collected in a string array.
    * The returned text will NOT include the start and end tags.
    * A zero length array is returned if no elements are found.
    *
    * @param   in           The Reader to get the XML from.
    * @param   tag          The tag to look for.
    * @return               An array of the elements found.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String[] getElements(java.io.Reader in, String tag) throws XmlHandlerException {
    java.util.Vector v = new java.util.Vector();
    try {
      while (true)
        v.addElement(getElement(in, tag));
    } catch (XmlValueNotFoundException e)  {
      // This will always happen, eventually no more are found
    }
    return getStringArray(v);
  }


  /**
    * Allows the text to be taken from a <code>java.io.Reader</code>.
    * Multiple instances of the element are collected in a string array.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    * A zero length array is returned if no elements are found.
    *
    * @param   in           The Reader to get the XML from.
    * @param   tag          The tag to look for.
    * @return               An array of the elements found or an empty
    *                       array otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String[] getElementsWithTags(java.io.Reader in, String tag) throws XmlHandlerException {
    java.util.Vector v = new java.util.Vector();
    try {
      while (true)
        v.addElement(getElementWithTags(in, tag));
    } catch (XmlValueNotFoundException e)  {
      // This will always happen, eventually no more are found
    }
    return getStringArray(v);
  }



  /**
    * Allows the text to be taken from a file.
    * Multiple instances of the element are collected in a string array.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    * A zero length array is returned if no elements are found.
    *
    * @param   in           The Reader to get the XML from.
    * @param   tag          The tag to look for.
    * @return               An array of the elements found.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  IOException  If the file cannot be opened.
    */
  public static String[] getElementsWithTagsFromFile(String filename, String tag)
                   throws java.io.IOException, XmlHandlerException {
    java.io.BufferedReader in
            = new java.io.BufferedReader(new java.io.FileReader(filename));
    String[] ss = getElementsWithTags(in, tag);
    in.close();
    return ss;
  }


  /**
    * Allows the text to be taken from a file.
    * The first instance of the element is returned as a string.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    *
    * @param   in           The Reader to get the XML from.
    * @param   tag          The tag to look for.
    * @return               An array of the elements found.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the element is not found.
    * @exception  IOException  If the file cannot be opened.
    */
  public static String getElementWithTagsFromFile(String filename, String tag)
                   throws java.io.IOException, XmlHandlerException {
    java.io.BufferedReader in
            = new java.io.BufferedReader(new java.io.FileReader(filename));
    String s = getElementWithTags(in, tag, "");
    in.close();
    return s;
  }











 /**
    * Allows the text to be taken from an XML element.
    * For example, from the string
    * <code>&lt;tag myname="andy"&gt;Some text&lt;/tag&gt;</code>
    * "Some text" will be returned using "tag" as the tag parameter.
    * All text between the start and end tags, including any contained
    * elements, will be returned. The tags can be anywhere in the string,
    * but the start tag must be before the end tag and both must be present.
    * In the event of any problems the empty string is returned.
    *
    * @param   in          The java.io.Reader to find the XML element in.
    * @param   tag         The tag to look for (without &lt; and &gt;).
    * @return              The element found or the empty string ("") otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the element is not found.
    */
  public static String getElement(java.io.Reader in, String tag)
           throws XmlHandlerException {
    return stringWithoutTags(getElementWithTags(in, tag));
  }


  /**
    * Allows the text to be taken from an XML element.
    *
    * @param   in          The java.io.Reader to find the XML element in.
    * @param   tag          The tag to look for.
    * @param   defaultValue The default value.
    * @return               The element found or the default otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String getElement(java.io.Reader in, String tag,
                                          String defaultValue)
                        throws XmlHandlerException {
    try {
      return getElement(in, tag);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }


  /**
    * Allows the text to be taken from an XML element.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    *
    * @param   in          The java.io.Reader to find the XML element in.
    * @param   tag          The tag to look for.
    * @param   defaultValue The default value.
    * @return               The element found or the default otherwise.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    */
  public static String getElementWithTags(java.io.Reader in, String tag,
                                          String defaultValue)
                        throws XmlHandlerException {
    try {
      return getElementWithTags(in, tag);
    } catch (XmlValueNotFoundException e) {
      return defaultValue;
    }
  }




  // This is the method used by all getElementXXX methods

  /**
    * Extracts an XML element from the specified Reader.
    * The returned text will include the start and end tags. This is important
    * if the values from name-value pairs are required.
    *
    * @param   in           The XML stream to examine.
    * @param   tag          The tag to look for.
    * @return               The element found.
    * @exception  XmlHandlerException  If there is a problem with the XML string.
    * @exception  XmlValueNotFoundException  If the element is not found.
    * @exception  NullPointerException If the tag or in is null.
    * @exception  IllegalArgumentException If the tag is zero length.
    */
  public static String getElementWithTags(java.io.Reader in, String tag)
           throws XmlHandlerException {
    if (tag == null)
      throw new java.lang.NullPointerException("tag = null in XmlHandler");
    if (in == null)
      throw new java.lang.NullPointerException("in = null in XmlHandler");
    if (tag.length() == 0)
      throw new java.lang.IllegalArgumentException("Zero length tag found by XmlHandler");

    char c = ' ';

      while (true) {                          // while... looking for start
        while (read(in, "No start tag for " + tag, true) != '<')        // search for <
          ;

        c = read(in, "No start tag for " + tag, true);                   // check next char

            // This may be a comment rather than a tag; if so search for end of comment
        if (c == '!') {                                            // looks like a comment
          if ((c = read(in, "No start tag for " + tag, true)) != '-')
            continue;  // not comment
          if ((c = read(in, "No start tag for " + tag, true)) != '-')
            continue;  // not comment
          // it is a comment
          findCommentEnd(in, tag);
          continue;
        }

      // up to "...<?"

        if (c != tag.charAt(0))
          continue;
        boolean flag = false;
        for (int i = 1; i < tag.length(); i++) {
          if ((c = read(in, "No start tag for " + tag, true)) != tag.charAt(i)) {
            flag = true;
            continue;           // have to continue out of two loops so use flag!
          }
        }
        if (flag) continue;
        if (((c = read(in, "No start tag for " + tag, true)) == '>') || Character.isWhitespace(c) || (c == '/'))
          break;
      }                                  // end while... looking for start

      StringBuffer sb = new StringBuffer("<" + tag + c);

      // up to "...<tagname?"
      while (c != '>') {       // search for <
        c = read(in, "No start tag for " + tag, true);
        sb.append(c);
      }
      if (sb.charAt(sb.length() - 2) == '/')    // empty element <tag/>
        return sb.toString().substring(0, sb.length() - 2) + "></" + tag + ">";

      while (true) {
        while ((c = read(in, "No end tag found for " + tag, false)) != '<')
          sb.append(c);

        c = read(in, "No start tag for " + tag, false);                   // check next char
        if (c == '!') {                                            // looks like a comment
          if ((c = read(in, "No start tag for " + tag, true)) != '-') {
            sb.append("<!" + c);
            continue;
          }
          if ((c = read(in, "No start tag for " + tag, true)) != '-') {
            sb.append("<!-" + c);
            continue;
          }
          // it is a comment
          findCommentEnd(in, tag);
          continue;
        }

        if (c != '/') {                                            // not end tag
          sb.append("<" + c);
          continue;
        }

        sb.append("</");
        String endTag = tag + ">";
        boolean flag = true;
        for (int i = 0; i < endTag.length(); i++) {
          sb.append((c = read(in, "No end tag for " + tag, false)));
          if (c != endTag.charAt(i)) {
            flag = false;
            break;
          }
        }
        if (flag) {
          break;
        }
      }

    return stringWithWhitespaceCollapsed(sb.toString());

  }









  /**
    * Adds the XML elements from a given <code>java.lang.String</code> to the given
    * <code>java.util.Hashtable</code>. The XML string is searched for each tag in an
    * array (which can be in any order). Entries are added with the tag as the name and the element
    * text (without tags) as the value. A zero-length string is added for tags that are not found.
    * Any entries already in the hashtable will be overwritten if they have the same names as one of
    * the tags, or unaffected otherwise.
    *
    * @param   xmlString     The XML string.
    * @param   tagNames      An array of tags to look for.
    * @param   h             The <code>java.util.Hashtable</code> that the name-value pairs will be put in.
    */
  public static void toHashtable(String xmlString, String[] tagNames, java.util.Hashtable h) throws XmlHandlerException {
    for (int i = 0; i < tagNames.length; i++)
      h.put(tagNames[i], XmlHandler.getElement(xmlString, tagNames[i], ""));
  }

  /**
    * Adds the XML elements from a given <code>java.lang.String</code> to the given
    * <code>java.util.Hashtable</code>. The XML string is searched for each tag in an
    * array (which can be in any order). Entries are added with the tag as the name and the element
    * text with tags as the value. A zero-length string is added for tags that are not found.
    * Any entries already in the hashtable will be overwritten if they have the same names as one of
    * the tags, or unaffected otherwise.
    *
    * @param   xmlString     The XML string.
    * @param   tagNames      An array of tags to look for.
    * @param   h             The <code>java.util.Hashtable</code> that the name-value pairs will be put in.
    */
  public static void toHashtableWithTags(String xmlString, String[] tagNames, java.util.Hashtable h) throws XmlHandlerException {
    for (int i = 0; i < tagNames.length; i++)
      h.put(tagNames[i], XmlHandler.getElementWithTags(xmlString, tagNames[i], ""));
  }

  /**
    * Adds the XML elements from a given <code>java.lang.String</code> to a new
    * <code>java.util.Hashtable</code>. The XML string is searched for each tag in an
    * array (which can be in any order). Entries are added with the tag as the name and the element
    * text (without tags) as the value. A zero-length string is added for tags that are not found.
    *
    * @param   xmlString     The XML string.
    * @param   tagNames      An array of tags to look for.
    * @return             The <code>java.util.Hashtable</code> with the name-value pairs in it.
    */
  public static java.util.Hashtable getElementsAsHashtable(String xmlString, String[] tagNames) throws XmlHandlerException {
    java.util.Hashtable h = new java.util.Hashtable();
    for (int i = 0; i < tagNames.length; i++)
      h.put(tagNames[i], XmlHandler.getElement(xmlString, tagNames[i], ""));
    return h;
  }

  /**
    * Adds the XML elements from a given <code>java.lang.String</code> to a new
    * <code>java.util.Hashtable</code>. The XML string is searched for each tag in an
    * array (which can be in any order). Entries are added with the tag as the name and the element
    * text with tags as the value. A zero-length string is added for tags that are not found.
    *
    * @param   xmlString     The XML string.
    * @param   tagNames      An array of tags to look for.
    * @return             The <code>java.util.Hashtable</code> with the name-value pairs in it.
    */
  public static java.util.Hashtable getElementsWithTagsAsHashtable(String xmlString, String[] tagNames) throws XmlHandlerException {
    java.util.Hashtable h = new java.util.Hashtable();
    for (int i = 0; i < tagNames.length; i++)
      h.put(tagNames[i], XmlHandler.getElementWithTags(xmlString, tagNames[i], ""));
    return h;
  }

  public static String toXml(java.util.Hashtable h) {
    return toXml(h, null);
  }

  public static String toXml(java.util.Hashtable h, String name) {
    StringBuffer sb = new StringBuffer();
    if (name != null)
      sb.append("<" + name + ">\n");
    java.util.Enumeration keys = h.keys();
    java.util.Enumeration elements = h.elements();
    Object key;
    while (keys.hasMoreElements()) {
      key = keys.nextElement();
      sb.append("<" + key + ">" + elements.nextElement() + "</" + key + ">\n");
    }
    if (name != null)
      sb.append("</" + name + ">\n");
    return sb.toString();
  }




    public static int getRanking(XmlUser xs, String term) {
        int res, n = xs.toString().indexOf(term);
        if (n != -1) {
            int m = xs.toString().length() - term.length();
            res = (100 - n) * (100 - m) / 100;
        } else {
            String s = stringWithoutAllTags(xs.toXml());
            n = s.indexOf(term);
            if (n != -1) {
                int m = s.length() - term.length();
                res = (100 - n) * (100 - m) / 150;
            } else {
                res = 0;
            }
        }
        return res;
    }

    /**
     * Takes the given string and swaps out any XML element with the name
     * "token", replacing it with the text provided by TokenProvider. Thus
     * <pre>
     * <tag>
     *   My favourite animal is the &lt;token value="animal"/>
     *   because &lt;token value="reason"/>.
     * </tag>
     * </pre>
     * ... could become...
     * <pre>
     * <tag>
     *   My favourite animal is the elephant
     *   because it is so big.
     * </tag>
     * </pre>
     *
     * @param s      The source string
     * @param tp     The TokenProvider object that will provide the text for swapping
     * @return       The substituted string
     * @throws XmlHandlerException If bad XML is encountered or the TokenProvider throws an exception.
     */
    public static String tokenSwap(String s, TokenProvider tp) throws XmlHandlerException {
        int start = 0, pos = findElement(s, "token", start);
        if (pos == -1) {
            if (s.indexOf("<token>") == -1) {
                return s;
            } else {
                throw new XmlValueNotFoundException("no value found for token in: " + s);
            }
        }

        StringBuffer sb = new StringBuffer();

        while (pos != -1) {
            sb.append(s.substring(start, pos));
            String tokenElement = XmlHandler.getElementWithTags(s.substring(pos), "token");
            String token = XmlHandler.getValue(tokenElement, "value");
            sb.append(tp.getTokenText(token));

            int end1 = s.indexOf("/>", pos);
            int end2 = s.indexOf("</token>", pos);
            if ((end1 == -1) && (end2 == -1)) {
                throw new XmlHandlerException("No terminator for token element in " + s,
                                 XmlHandlerException.BAD_XML);
            }
            if (end1 == -1) {
                start = end2 + 8;
            } else if (end2 == -1) {
                start = end1 + 2;
            } else if (end1 > end2) {
                start = end2 + 8;
            } else {
                start = end1 + 2;
            }
            pos = findElement(s, "token", start);
        }

        sb.append(s.substring(start));
        return sb.toString();
    }

    /*
     * Finds the position of the '<' character that starts the XML
     * element with the given tag, searching from the given position.
     * Takes into account the character after the tag, to ensure it has the
     * whole tag (succeeding chacter must be one of '>', ' ', '\t' or '\n').
     * Tested, but not with JUnit.
     */
    private static int findElement(String s, String tag, int pos) {
        int pos1 = s.indexOf("<" + tag + ">", pos);
        int pos2 = s.indexOf("<" + tag + " ", pos);
        int pos3 = s.indexOf("<" + tag + "\t", pos);
        int pos4 = s.indexOf("<" + tag + "\n", pos);

        int pos5 = (pos1 == -1 ? pos2 : (pos2 == -1 ? pos1 : (pos1 > pos2 ? pos2 : pos1)));
        int pos6 = (pos3 == -1 ? pos4 : (pos4 == -1 ? pos3 : (pos3 > pos4 ? pos4 : pos3)));
        return (pos5 == -1 ? pos6 : (pos6 == -1 ? pos5 : (pos5 > pos6 ? pos6 : pos5)));
    }

    /**
     * Used by tokenSwap to substitute tokens for text.
     */
    public interface TokenProvider {
        /**
         * Gets the appropriate text for the given token. If the token is not
         * recognised, a default value, which could be the empty string, should
         * be returned, or an XmlValueNotFoundException thrown. Null should not be returned.
         *
         * @param token  An identifying string.
         * @return The replacement text.
         * @throws XmlValueNotFoundException if the token is not recognised and no default
         *                 is offered.
         */
        public String getTokenText(String token) throws XmlValueNotFoundException;
    }

    /**
     * Converts a string into something suitable for an XML name, so, for example,
     * "My Favourite Color" becomes "my-favourite-color".
     *
     * @param s  String to be converted.
     * @return   A string in lowercase with all whitespace converted to dashes.
     */
    public static String getSafe(String s) {
        java.util.StringTokenizer st = new java.util.StringTokenizer(s, ",. \n\t\r");
        if (!st.hasMoreTokens())
            throw new IllegalArgumentException("Invalid argument to getSafe (" + s + ")");
        StringBuffer sb = new StringBuffer(st.nextToken());
        while (st.hasMoreTokens())
            sb.append("-" + st.nextToken());
        return sb.toString().toLowerCase();
    }




// --------------------- Some private methods ----------------------


   // Method takes characters from the java.io.Reader in, until it finds "-->"

  private static void findCommentEnd(java.io.Reader in, String tag) throws XmlHandlerException {
    while (true) {             // search for -->
      while (read(in, "No end to comment for tag " + tag, false) != '-')    // find -
        ;
      if (read(in, "No end to comment for tag " + tag, false) != '-')
        continue;              //found -? not --, so keep loking for -->
      if (read(in, "No end to comment for tag " + tag, false) == '>')
        break;                 //found -->, so break out of comment loop
    }
  }


  private static String stringWithoutComments(String s) {
    return Utils.deleteSection(s, "<!--", "-->");
  }


  private static char read(java.io.Reader in, String errorMsg, boolean notFoundError)
                  throws XmlHandlerException  {
    try {
      int c = in.read();
      if (c == -1)
        throw (notFoundError ?
                           new XmlValueNotFoundException(errorMsg) :
                           new XmlHandlerException(errorMsg, XmlHandlerException.BAD_XML));
      else
         return (char)c;
    } catch (java.io.IOException e) {
        throw (notFoundError ?
                           new XmlValueNotFoundException(errorMsg) :
                           new XmlHandlerException(errorMsg, XmlHandlerException.BAD_XML));
    }
  }

  public static String stringWithoutAllTags(String s) {
    StringBuffer sb = new StringBuffer();
    boolean inTag = false;
    for (int i = 0; i < s.length(); i++) {
      if (inTag) {
        if (s.charAt(i) == '>') {
          inTag = false;
        }
      } else {
        if (s.charAt(i) == '<') {
          inTag = true;
          sb.append(" ");
        } else {
          sb.append(s.charAt(i));
        }
      }
    }
    return sb.toString();
  }




    public static XmlUser rebuild(String xmlString, Class myClass)
                throws XmlHandlerException {
        String s = XmlHandler.getValue(xmlString, "class", "");
        XmlUser item = null;
        try {
            if (s.length() == 0) {
                item = (XmlUser) myClass.newInstance();
            } else {
                item = (XmlUser) Class.forName(s).newInstance();
            }
        } catch (ClassNotFoundException ex) {
            throw new XmlHandlerException("ClassNotFoundException in XmlDatatbase.add: " + ex, XmlHandlerException.BAD_INSTANTIATION);
        } catch (IllegalAccessException ex) {
            throw new XmlHandlerException("IllegalAccessException in XmlDatatbase.add: " + ex, XmlHandlerException.BAD_INSTANTIATION);
        } catch (InstantiationException ex) {
            throw new XmlHandlerException("InstantiationException in XmlDatatbase.add: " + ex, XmlHandlerException.BAD_INSTANTIATION);
        }
        item.fromXml(xmlString);
        return item;
    }



  private static String stringWithoutTags(String s) {
    String trimmed = s.trim();
    // Check for empty element <tag/> or <tag />
    if ((trimmed.lastIndexOf('<') == 0) && (trimmed.indexOf('>') == (trimmed.length() - 1)))
      return "";
    return trimmed.substring(trimmed.indexOf('>') + 1, trimmed.lastIndexOf('<')).trim();
  }

  private static String stringWithWhitespaceCollapsed(String s) {
    java.util.StringTokenizer st = new java.util.StringTokenizer(s);
    if (!st.hasMoreTokens())
      return "";
    StringBuffer sb = new StringBuffer(st.nextToken());
    while (st.hasMoreTokens()) {
      sb.append(" " + st.nextToken());
    }
    return sb.toString();
  }

  private static String[] getStringArray(java.util.Vector v) {
    String ss[] = new String[v.size()];
    for (int i = 0; i < v.size(); i++)
      ss[i] = v.elementAt(i).toString();
    return ss;
  }


}

